import sqlite3
from unittest import TestCase

import pytest

from piccolo.columns import Integer, Varchar
from piccolo.query.methods.insert import OnConflictAction
from piccolo.table import Table
from piccolo.utils.lazy_loader import LazyLoader
from tests.base import (
    DBTestCase,
    engine_version_lt,
    engines_only,
    is_running_sqlite,
)
from tests.example_apps.music.tables import Band, Manager

asyncpg = LazyLoader("asyncpg", globals(), "asyncpg")


class TestInsert(DBTestCase):
    def test_insert(self):
        self.insert_rows()

        Band.insert(Band(name="Rustaceans", popularity=100)).run_sync()

        response = Band.select(Band.name).run_sync()
        names = [i["name"] for i in response]

        self.assertIn("Rustaceans", names)

    def test_add(self):
        self.insert_rows()

        Band.insert().add(Band(name="Rustaceans", popularity=100)).run_sync()

        response = Band.select(Band.name).run_sync()
        names = [i["name"] for i in response]

        self.assertIn("Rustaceans", names)

    def test_incompatible_type(self):
        """
        You shouldn't be able to add instances of a different table.
        """
        with self.assertRaises(TypeError):
            Band.insert().add(Manager(name="Guido")).run_sync()

    def test_insert_curly_braces(self):
        """
        You should be able to insert curly braces without an error.
        """
        self.insert_rows()

        Band.insert(Band(name="{}", popularity=100)).run_sync()

        response = Band.select(Band.name).run_sync()
        names = [i["name"] for i in response]

        self.assertIn("{}", names)

    @pytest.mark.skipif(
        is_running_sqlite() and engine_version_lt(3.35),
        reason="SQLite version not supported",
    )
    def test_insert_returning(self):
        """
        Make sure update works with the `returning` clause.
        """
        response = (
            Manager.insert(Manager(name="Maz"))
            .returning(Manager.name)
            .run_sync()
        )

        self.assertListEqual(response, [{"name": "Maz"}])

    @pytest.mark.skipif(
        is_running_sqlite() and engine_version_lt(3.35),
        reason="SQLite version not supported",
    )
    def test_insert_returning_alias(self):
        """
        Make sure update works with the `returning` clause.
        """
        response = (
            Manager.insert(Manager(name="Maz"))
            .returning(Manager.name.as_alias("manager_name"))
            .run_sync()
        )

        self.assertListEqual(response, [{"manager_name": "Maz"}])


@pytest.mark.skipif(
    is_running_sqlite() and engine_version_lt(3.24),
    reason="SQLite version not supported",
)
class TestOnConflict(TestCase):
    class Band(Table):
        name = Varchar(unique=True)
        popularity = Integer()

    def setUp(self) -> None:
        Band = self.Band
        Band.create_table().run_sync()
        self.band = Band({Band.name: "Pythonistas", Band.popularity: 1000})
        self.band.save().run_sync()

    def tearDown(self) -> None:
        Band = self.Band
        Band.alter().drop_table().run_sync()

    def test_do_update(self):
        """
        Make sure that `DO UPDATE` works.
        """
        Band = self.Band

        new_popularity = self.band.popularity + 1000

        Band.insert(
            Band(name=self.band.name, popularity=new_popularity)
        ).on_conflict(
            target=Band.name,
            action="DO UPDATE",
            values=[Band.popularity],
        ).run_sync()

        self.assertListEqual(
            Band.select().run_sync(),
            [
                {
                    "id": self.band.id,
                    "name": self.band.name,
                    "popularity": new_popularity,  # changed
                }
            ],
        )

    def test_do_update_tuple_values(self):
        """
        Make sure we can use tuples in ``values``.
        """
        Band = self.Band

        new_popularity = self.band.popularity + 1000
        new_name = "Rustaceans"

        Band.insert(
            Band(
                id=self.band.id,
                name=new_name,
                popularity=new_popularity,
            )
        ).on_conflict(
            action="DO UPDATE",
            target=Band.id,
            values=[
                (Band.name, new_name),
                (Band.popularity, new_popularity + 2000),
            ],
        ).run_sync()

        self.assertListEqual(
            Band.select().run_sync(),
            [
                {
                    "id": self.band.id,
                    "name": new_name,
                    "popularity": new_popularity + 2000,
                }
            ],
        )

    def test_do_update_no_values(self):
        """
        Make sure that `DO UPDATE` with no `values` raises an exception.
        """
        Band = self.Band

        new_popularity = self.band.popularity + 1000

        with self.assertRaises(ValueError) as manager:
            Band.insert(
                Band(name=self.band.name, popularity=new_popularity)
            ).on_conflict(
                target=Band.name,
                action="DO UPDATE",
            ).run_sync()

        self.assertEqual(
            manager.exception.__str__(),
            "No values specified for `on conflict`",
        )

    @engines_only("postgres", "cockroach")
    def test_target_tuple(self):
        """
        Make sure that a composite unique constraint can be used as a target.

        We only run it on Postgres and Cockroach because we use ALTER TABLE
        to add a contraint, which SQLite doesn't support.
        """
        Band = self.Band

        # Add a composite unique constraint:
        Band.raw(
            "ALTER TABLE band ADD CONSTRAINT id_name_unique UNIQUE (id, name)"
        ).run_sync()

        Band.insert(
            Band(
                id=self.band.id,
                name=self.band.name,
                popularity=self.band.popularity,
            )
        ).on_conflict(
            target=(Band.id, Band.name),
            action="DO NOTHING",
        ).run_sync()

    @engines_only("postgres", "cockroach")
    def test_target_string(self):
        """
        Make sure we can explicitly specify the name of target constraint using
        a string.

        We just test this on Postgres for now, as we have to get the constraint
        name from the database.
        """
        Band = self.Band

        constraint_name = Band.raw(
            """
            SELECT constraint_name
            FROM information_schema.constraint_column_usage
            WHERE column_name = 'name'
                AND table_name = 'band';
            """
        ).run_sync()[0]["constraint_name"]

        query = Band.insert(Band(name=self.band.name)).on_conflict(
            target=constraint_name,
            action="DO NOTHING",
        )
        self.assertIn(f'ON CONSTRAINT "{constraint_name}"', query.__str__())
        query.run_sync()

    def test_violate_non_target(self):
        """
        Make sure that if we specify a target constraint, but violate a
        different constraint, then we still get the error.
        """
        Band = self.Band

        new_popularity = self.band.popularity + 1000

        with self.assertRaises(Exception) as manager:
            Band.insert(
                Band(name=self.band.name, popularity=new_popularity)
            ).on_conflict(
                target=Band.id,  # Target the primary key instead.
                action="DO UPDATE",
                values=[Band.popularity],
            ).run_sync()

        if self.Band._meta.db.engine_type in ("postgres", "cockroach"):
            self.assertIsInstance(
                manager.exception, asyncpg.exceptions.UniqueViolationError
            )
        elif self.Band._meta.db.engine_type == "sqlite":
            self.assertIsInstance(manager.exception, sqlite3.IntegrityError)

    def test_where(self):
        """
        Make sure we can pass in a `where` argument.
        """
        Band = self.Band

        new_popularity = self.band.popularity + 1000

        query = Band.insert(
            Band(name=self.band.name, popularity=new_popularity)
        ).on_conflict(
            target=Band.name,
            action="DO UPDATE",
            values=[Band.popularity],
            where=Band.popularity < self.band.popularity,
        )

        self.assertIn(
            f'WHERE "band"."popularity" < {self.band.popularity}',
            query.__str__(),
        )

        query.run_sync()

    def test_do_nothing_where(self):
        """
        Make sure an error is raised if `where` is used with `DO NOTHING`.
        """
        Band = self.Band

        with self.assertRaises(ValueError) as manager:
            Band.insert(Band()).on_conflict(
                action="DO NOTHING",
                where=Band.popularity < self.band.popularity,
            )

        self.assertEqual(
            manager.exception.__str__(),
            "The `where` option can only be used with DO NOTHING.",
        )

    def test_do_nothing(self):
        """
        Make sure that `DO NOTHING` works.
        """
        Band = self.Band

        new_popularity = self.band.popularity + 1000

        Band.insert(
            Band(name="Pythonistas", popularity=new_popularity)
        ).on_conflict(action="DO NOTHING").run_sync()

        self.assertListEqual(
            Band.select().run_sync(),
            [
                {
                    "id": self.band.id,
                    "name": self.band.name,
                    "popularity": self.band.popularity,
                }
            ],
        )

    @engines_only("sqlite")
    def test_multiple_do_update(self):
        """
        Make sure multiple `ON CONFLICT` clauses work for SQLite.
        """
        Band = self.Band

        new_popularity = self.band.popularity + 1000

        # Conflicting with name - should update.
        Band.insert(
            Band(name="Pythonistas", popularity=new_popularity)
        ).on_conflict(action="DO NOTHING", target=Band.id).on_conflict(
            action="DO UPDATE", target=Band.name, values=[Band.popularity]
        ).run_sync()

        self.assertListEqual(
            Band.select().run_sync(),
            [
                {
                    "id": self.band.id,
                    "name": self.band.name,
                    "popularity": new_popularity,  # changed
                }
            ],
        )

    @engines_only("sqlite")
    def test_multiple_do_nothing(self):
        """
        Make sure multiple `ON CONFLICT` clauses work for SQLite.
        """
        Band = self.Band

        new_popularity = self.band.popularity + 1000

        # Conflicting with ID - should be ignored.
        Band.insert(
            Band(
                id=self.band.id,
                name="Pythonistas",
                popularity=new_popularity,
            )
        ).on_conflict(action="DO NOTHING", target=Band.id).on_conflict(
            action="DO UPDATE",
            target=Band.name,
            values=[Band.popularity],
        ).run_sync()

        self.assertListEqual(
            Band.select().run_sync(),
            [
                {
                    "id": self.band.id,
                    "name": self.band.name,
                    "popularity": self.band.popularity,
                }
            ],
        )

    @engines_only("postgres", "cockroach")
    def test_mutiple_error(self):
        """
        Postgres and Cockroach don't support multiple `ON CONFLICT` clauses.
        """
        with self.assertRaises(NotImplementedError) as manager:
            Band = self.Band

            Band.insert(Band()).on_conflict(action="DO NOTHING").on_conflict(
                action="DO UPDATE",
            ).run_sync()

        assert manager.exception.__str__() == (
            "Postgres and Cockroach only support a single ON CONFLICT clause."
        )

    def test_all_columns(self):
        """
        We can use ``all_columns`` instead of specifying the ``values``
        manually.
        """
        Band = self.Band

        new_popularity = self.band.popularity + 1000
        new_name = "Rustaceans"

        # Conflicting with ID - should be ignored.
        q = Band.insert(
            Band(
                id=self.band.id,
                name=new_name,
                popularity=new_popularity,
            )
        ).on_conflict(
            action="DO UPDATE",
            target=Band.id,
            values=Band.all_columns(),
        )
        q.run_sync()

        self.assertListEqual(
            Band.select().run_sync(),
            [
                {
                    "id": self.band.id,
                    "name": new_name,
                    "popularity": new_popularity,
                }
            ],
        )

    def test_enum(self):
        """
        A string literal can be passed in, or an enum, to determine the action.
        Make sure that the enum works.
        """
        Band = self.Band

        Band.insert(
            Band(
                id=self.band.id,
                name=self.band.name,
                popularity=self.band.popularity,
            )
        ).on_conflict(action=OnConflictAction.do_nothing).run_sync()

        self.assertListEqual(
            Band.select().run_sync(),
            [
                {
                    "id": self.band.id,
                    "name": self.band.name,
                    "popularity": self.band.popularity,
                }
            ],
        )



from piccolo.apps.migrations.auto import MigrationManager

ID = "2020-03-31T20:38:22"


async def forwards():
    manager = MigrationManager(migration_id=ID)
    manager.add_table("Band", tablename="band")
    manager.add_column(
        table_class_name="Band",
        tablename="band",
        column_name="name",
        column_class_name="Varchar",
        params={
            "length": 150,
            "default": "",
            "null": True,
            "primary": False,
            "key": False,
            "unique": False,
            "index": False,
        },
    )

    return manager



from piccolo.conf.apps import AppConfig, Command

from .commands.backwards import backwards
from .commands.check import check
from .commands.clean import clean
from .commands.forwards import forwards
from .commands.new import new

APP_CONFIG = AppConfig(
    app_name="migrations",
    migrations_folder_path="",
    commands=[
        Command(callable=backwards, aliases=["b", "back", "backward"]),
        Command(callable=check),
        Command(callable=clean),
        Command(callable=forwards, aliases=["f", "forward"]),
        Command(callable=new, aliases=["n", "create"]),
    ],
)



from __future__ import annotations

import typing as t

from piccolo.query.base import DDL
from piccolo.query.methods.create_index import CreateIndex

if t.TYPE_CHECKING:  # pragma: no cover
    from piccolo.table import Table


class Create(DDL):
    """
    Creates a database table.
    """

    __slots__ = ("if_not_exists", "only_default_columns", "auto_create_schema")

    def __init__(
        self,
        table: t.Type[Table],
        if_not_exists: bool = False,
        only_default_columns: bool = False,
        auto_create_schema: bool = True,
        **kwargs,
    ):
        """
        :param table:
            The table to create.
        :param if_not_exists:
            If ``True``, no error will be raised if this table already exists.
        :param only_default_columns:
            If ``True``, just the basic table and default primary key are
            created, rather than all columns. Not typically needed.
        :param auto_create_schema:
            If the table belongs to a database schema, then make sure the
            schema exists before creating the table.

        """
        super().__init__(table, **kwargs)
        self.if_not_exists = if_not_exists
        self.only_default_columns = only_default_columns
        self.auto_create_schema = auto_create_schema

    @property
    def default_ddl(self) -> t.Sequence[str]:
        ddl: t.List[str] = []

        schema_name = self.table._meta.schema
        if (
            self.auto_create_schema
            and schema_name is not None
            and schema_name != "public"
            and self.engine_type != "sqlite"
        ):
            from piccolo.schema import CreateSchema

            ddl.append(
                CreateSchema(
                    schema_name=schema_name,
                    if_not_exists=True,
                    db=self.table._meta.db,
                ).ddl
            )

        prefix = "CREATE TABLE"
        if self.if_not_exists:
            prefix += " IF NOT EXISTS"

        if self.only_default_columns:
            columns = self.table._meta.non_default_columns
        else:
            columns = self.table._meta.columns

        base = f"{prefix} {self.table._meta.get_formatted_tablename()}"
        columns_sql = ", ".join(i.ddl for i in columns)
        ddl.append(f"{base} ({columns_sql})")

        for column in columns:
            if column._meta.index is True:
                ddl.extend(
                    CreateIndex(
                        table=self.table,
                        columns=[column],
                        method=column._meta.index_method,
                        if_not_exists=self.if_not_exists,
                    ).ddl
                )

        return ddl



from __future__ import annotations

import typing as t
from dataclasses import dataclass

from piccolo.utils.sync import run_sync

if t.TYPE_CHECKING:  # pragma: no cover
    from piccolo.columns import Column
    from piccolo.table import Table


@dataclass
class Refresh:
    """
    Used to refresh :class:`Table <piccolo.table.Table>` instances with the
    latest data data from the database. Accessible via
    :meth:`refresh <piccolo.table.Table.refresh>`.

    :param instance:
        The instance to refresh.
    :param columns:
        Which columns to refresh - it not specified, then all columns are
        refreshed.

    """

    instance: Table
    columns: t.Optional[t.Sequence[Column]] = None

    @property
    def _columns(self) -> t.Sequence[Column]:
        """
        Works out which columns the user wants to refresh.
        """
        if self.columns:
            return self.columns

        return [
            i for i in self.instance._meta.columns if not i._meta.primary_key
        ]

    async def run(
        self, in_pool: bool = True, node: t.Optional[str] = None
    ) -> Table:
        """
        Run it asynchronously. For example::

            await my_instance.refresh().run()

            # or for convenience:
            await my_instance.refresh()

        Modifies the instance in place, but also returns it as a convenience.

        """

        instance = self.instance

        if not instance._exists_in_db:
            raise ValueError("The instance doesn't exist in the database.")

        pk_column = instance._meta.primary_key

        primary_key_value = getattr(instance, pk_column._meta.name, None)

        if primary_key_value is None:
            raise ValueError("The instance's primary key value isn't defined.")

        columns = self._columns
        if not columns:
            raise ValueError("No columns to fetch.")

        updated_values = (
            await instance.__class__.select(*columns)
            .where(pk_column == primary_key_value)
            .first()
            .run(node=node, in_pool=in_pool)
        )

        if updated_values is None:
            raise ValueError(
                "The object doesn't exist in the database any more."
            )

        for key, value in updated_values.items():
            setattr(instance, key, value)

        return instance

    def __await__(self):
        """
        If the user doesn't explicity call :meth:`run`, proxy to it as a
        convenience.
        """
        return self.run().__await__()

    def run_sync(self, *args, **kwargs) -> Table:
        """
        Run it synchronously. For example::

            my_instance.refresh().run_sync()

        """
        return run_sync(self.run(*args, **kwargs))



from unittest import TestCase

from piccolo.columns.column_types import Real
from piccolo.table import Table


class MyTable(Table):
    column_a = Real()


class TestReal(TestCase):
    def setUp(self):
        MyTable.create_table().run_sync()

    def tearDown(self):
        MyTable.alter().drop_table().run_sync()

    def test_creation(self):
        row = MyTable(column_a=1.23)
        row.save().run_sync()

        _row = MyTable.objects().first().run_sync()
        self.assertEqual(type(_row.column_a), float)
        self.assertAlmostEqual(_row.column_a, 1.23)



import os

from piccolo.conf.apps import AppConfig, table_finder

CURRENT_DIRECTORY = os.path.dirname(os.path.abspath(__file__))


APP_CONFIG = AppConfig(
    app_name="music",
    table_classes=table_finder(["tests.example_apps.music.tables"]),
    migrations_folder_path=os.path.join(
        CURRENT_DIRECTORY, "piccolo_migrations"
    ),
    commands=[],
)



"""
This is a useful table when we want to test all possible column types.
"""

from piccolo.columns.column_types import (
    JSON,
    JSONB,
    UUID,
    BigInt,
    Boolean,
    Bytea,
    Date,
    DoublePrecision,
    ForeignKey,
    Integer,
    Interval,
    Numeric,
    Real,
    SmallInt,
    Text,
    Timestamp,
    Timestamptz,
    Varchar,
)
from piccolo.engine.finder import engine_finder
from piccolo.table import Table

engine = engine_finder()


class SmallTable(Table):
    varchar_col = Varchar()


if engine.engine_type != "cockroach":  # type: ignore

    class MegaTable(Table):  # type: ignore
        """
        A table containing all of the column types, different column kwargs.
        """

        bigint_col = BigInt()
        boolean_col = Boolean()
        bytea_col = Bytea()
        date_col = Date()
        foreignkey_col = ForeignKey(SmallTable)
        integer_col = Integer()
        interval_col = Interval()
        json_col = JSON()
        jsonb_col = JSONB()
        numeric_col = Numeric(digits=(5, 2))
        real_col = Real()
        double_precision_col = DoublePrecision()
        smallint_col = SmallInt()
        text_col = Text()
        timestamp_col = Timestamp()
        timestamptz_col = Timestamptz()
        uuid_col = UUID()
        varchar_col = Varchar()

        unique_col = Varchar(unique=True)
        null_col = Varchar(null=True)
        not_null_col = Varchar(null=False)

else:

    class MegaTable(Table):  # type: ignore
        """
        Special version for Cockroach.
        A table containing all of the column types, different column kwargs.
        """

        bigint_col = BigInt()
        boolean_col = Boolean()
        bytea_col = Bytea()
        date_col = Date()
        foreignkey_col = ForeignKey(SmallTable)
        integer_col = BigInt()
        interval_col = Interval()
        json_col = JSONB()
        jsonb_col = JSONB()
        numeric_col = Numeric(digits=(5, 2))
        real_col = Real()
        double_precision_col = DoublePrecision()
        smallint_col = SmallInt()
        text_col = Text()
        timestamp_col = Timestamp()
        timestamptz_col = Timestamptz()
        uuid_col = UUID()
        varchar_col = Varchar()

        unique_col = Varchar(unique=True)
        null_col = Varchar(null=True)
        not_null_col = Varchar(null=False)



import sys
import typing as t

from piccolo.conf.apps import AppConfig, AppRegistry, Finder
from piccolo.table import Table

try:
    import IPython  # type: ignore
    from IPython.core.interactiveshell import _asyncio_runner  # type: ignore

    IPYTHON = True
except ImportError:
    IPYTHON = False


def start_ipython_shell(
    **tables: t.Dict[str, t.Type[Table]]
):  # pragma: no cover
    if not IPYTHON:
        sys.exit(
            "Install iPython using `pip install ipython` to use this feature."
        )

    existing_global_names = globals().keys()
    for table_class_name, table_class in tables.items():
        if table_class_name not in existing_global_names:
            globals()[table_class_name] = table_class

    IPython.embed(using=_asyncio_runner, colors="neutral")


def run():
    """
    Runs an iPython shell, and automatically imports all of the Table classes
    from your project.
    """
    app_registry: AppRegistry = Finder().get_app_registry()

    tables = {}
    if app_registry.app_configs:
        spacer = "-------"

        print(spacer)

        for app_name, app_config in app_registry.app_configs.items():
            app_config: AppConfig = app_config
            print(f"Importing {app_name} tables:")
            if app_config.table_classes:
                for table_class in sorted(
                    app_config.table_classes, key=lambda x: x.__name__
                ):
                    table_class_name = table_class.__name__
                    print(f"- {table_class_name}")
                    tables[table_class_name] = table_class
            else:
                print("- None")

        print(spacer)

    start_ipython_shell(**tables)



import json
from unittest import TestCase

from tests.base import DBTestCase, engine_is
from tests.example_apps.music.tables import Band, RecordingStudio


class TestOutputList(DBTestCase):
    def test_output_as_list(self):
        self.insert_row()

        response = Band.select(Band.name).output(as_list=True).run_sync()
        self.assertEqual(response, ["Pythonistas"])

        # Make sure that if no rows are found, an empty list is returned.
        empty_response = (
            Band.select(Band.name)
            .where(Band.name == "ABC123")
            .output(as_list=True)
            .run_sync()
        )
        self.assertEqual(empty_response, [])


class TestOutputJSON(DBTestCase):
    def test_output_as_json(self):
        self.insert_row()

        response = Band.select(Band.name).output(as_json=True).run_sync()

        self.assertEqual(json.loads(response), [{"name": "Pythonistas"}])


class TestOutputLoadJSON(TestCase):
    def setUp(self):
        RecordingStudio.create_table().run_sync()

    def tearDown(self):
        RecordingStudio.alter().drop_table().run_sync()

    def test_select(self):
        json = {"a": 123}

        RecordingStudio(facilities=json, facilities_b=json).save().run_sync()

        results = RecordingStudio.select().output(load_json=True).run_sync()

        if engine_is("cockroach"):
            self.assertEqual(
                results,
                [
                    {
                        "id": results[0]["id"],
                        "facilities": {"a": 123},
                        "facilities_b": {"a": 123},
                    }
                ],
            )
        else:
            self.assertEqual(
                results,
                [
                    {
                        "id": 1,
                        "facilities": {"a": 123},
                        "facilities_b": {"a": 123},
                    }
                ],
            )

    def test_objects(self):
        json = {"a": 123}

        RecordingStudio(facilities=json, facilities_b=json).save().run_sync()

        results = RecordingStudio.objects().output(load_json=True).run_sync()

        self.assertEqual(results[0].facilities, json)
        self.assertEqual(results[0].facilities_b, json)


class TestOutputNested(DBTestCase):
    def test_output_nested(self):
        self.insert_row()

        response = (
            Band.select(Band.name, Band.manager.name)
            .output(nested=True)
            .run_sync()
        )
        self.assertEqual(
            response, [{"name": "Pythonistas", "manager": {"name": "Guido"}}]
        )

    def test_output_nested_with_first(self):
        self.insert_row()

        response = (
            Band.select(Band.name, Band.manager.name)
            .first()
            .output(nested=True)
            .run_sync()
        )
        self.assertDictEqual(
            response, {"name": "Pythonistas", "manager": {"name": "Guido"}}
        )



from unittest import TestCase

from piccolo.apps.migrations.tables import Migration


class TestMigrationTable(TestCase):
    def test_migration_table(self):
        Migration.create_table(if_not_exists=True).run_sync()
        Migration.select().run_sync()
        Migration.alter().drop_table().run_sync()



from __future__ import annotations

import asyncio
import collections.abc
import itertools
import typing as t
from dataclasses import dataclass, field
from enum import Enum, auto

from typing_extensions import Literal

from piccolo.columns import And, Column, Or, Where
from piccolo.columns.column_types import ForeignKey
from piccolo.custom_types import Combinable
from piccolo.querystring import QueryString
from piccolo.utils.list import flatten
from piccolo.utils.sql_values import convert_to_sql_value

if t.TYPE_CHECKING:  # pragma: no cover
    from piccolo.columns.base import Selectable
    from piccolo.table import Table  # noqa


class DistinctOnError(ValueError):
    """
    Raised when ``DISTINCT ON`` queries are malformed.
    """

    pass


@dataclass
class Distinct:
    __slots__ = ("enabled", "on")

    enabled: bool
    on: t.Optional[t.Sequence[Column]]

    @property
    def querystring(self) -> QueryString:
        if self.enabled:
            if self.on:
                column_names = ", ".join(
                    i._meta.get_full_name(with_alias=False) for i in self.on
                )
                return QueryString(f" DISTINCT ON ({column_names})")
            else:
                return QueryString(" DISTINCT")
        else:
            return QueryString(" ALL")

    def validate_on(self, order_by: OrderBy):
        """
        When using the `on` argument, the first column must match the first
        order by column.

        :raises DistinctOnError:
            If the columns don't match.

        """
        validated = True

        try:
            first_order_column = order_by.order_by_items[0].columns[0]
        except IndexError:
            validated = False
        else:
            if not self.on:
                validated = False
            elif isinstance(first_order_column, Column) and not self.on[
                0
            ]._equals(first_order_column):
                validated = False

        if not validated:
            raise DistinctOnError(
                "The first `order_by` column must match the first column "
                "passed to `on`."
            )

    def __str__(self) -> str:
        return self.querystring.__str__()

    def copy(self) -> Distinct:
        return self.__class__(enabled=self.enabled, on=self.on)


@dataclass
class Limit:
    __slots__ = ("number",)

    number: int

    def __post_init__(self):
        if type(self.number) != int:
            raise TypeError("Limit must be an integer")

    @property
    def querystring(self) -> QueryString:
        return QueryString(f" LIMIT {self.number}")

    def __str__(self) -> str:
        return self.querystring.__str__()

    def copy(self) -> Limit:
        return self.__class__(number=self.number)


@dataclass
class AsOf:
    __slots__ = ("interval",)

    interval: str

    def __post_init__(self):
        if type(self.interval) != str:
            raise TypeError("As Of must be a string. Example: '-1s'")

    @property
    def querystring(self) -> QueryString:
        return QueryString(f" AS OF SYSTEM TIME '{self.interval}'")

    def __str__(self) -> str:
        return self.querystring.__str__()


@dataclass
class Offset:
    __slots__ = ("number",)

    number: int

    def __post_init__(self):
        if type(self.number) != int:
            raise TypeError("Limit must be an integer")

    @property
    def querystring(self) -> QueryString:
        return QueryString(f" OFFSET {self.number}")

    def __str__(self) -> str:
        return self.querystring.__str__()


@dataclass
class OrderByRaw:
    __slots__ = ("sql",)

    sql: str


@dataclass
class OrderByItem:
    __slots__ = ("columns", "ascending")

    columns: t.Sequence[t.Union[Column, OrderByRaw]]
    ascending: bool


@dataclass
class OrderBy:
    order_by_items: t.List[OrderByItem] = field(default_factory=list)

    @property
    def querystring(self) -> QueryString:
        order_by_strings: t.List[str] = []
        for order_by_item in self.order_by_items:
            order = "ASC" if order_by_item.ascending else "DESC"
            for column in order_by_item.columns:
                if isinstance(column, Column):
                    expression = column._meta.get_full_name(with_alias=False)
                elif isinstance(column, OrderByRaw):
                    expression = column.sql
                else:
                    raise ValueError("Unrecognised order_by")

                order_by_strings.append(f"{expression} {order}")

        return QueryString(f" ORDER BY {', '.join(order_by_strings)}")

    def __str__(self):
        return self.querystring.__str__()


@dataclass
class Returning:
    __slots__ = ("columns",)

    columns: t.List[Column]

    @property
    def querystring(self) -> QueryString:
        column_names = []
        for column in self.columns:
            column_names.append(
                f'"{column._meta.db_column_name}" AS "{column._alias}"'
                if column._alias
                else f'"{column._meta.db_column_name}"'
            )

        columns_string = ", ".join(column_names)

        return QueryString(f" RETURNING {columns_string}")

    def __str__(self):
        return self.querystring.__str__()


@dataclass
class Output:

    as_json: bool = False
    as_list: bool = False
    as_objects: bool = False
    load_json: bool = False
    nested: bool = False

    def copy(self) -> Output:
        return self.__class__(
            as_json=self.as_json,
            as_list=self.as_list,
            as_objects=self.as_objects,
            load_json=self.load_json,
            nested=self.nested,
        )


class CallbackType(Enum):
    success = auto()


@dataclass
class Callback:
    kind: CallbackType
    target: t.Callable


@dataclass
class WhereDelegate:

    _where: t.Optional[Combinable] = None
    _where_columns: t.List[Column] = field(default_factory=list)

    def get_where_columns(self):
        """
        Retrieves all columns used in the where clause - in case joins are
        needed.
        """
        self._where_columns = []
        self._extract_columns(self._where)
        return self._where_columns

    def _extract_columns(self, combinable: Combinable):
        if isinstance(combinable, Where):
            self._where_columns.append(combinable.column)
        elif isinstance(combinable, (And, Or)):
            self._extract_columns(combinable.first)
            self._extract_columns(combinable.second)

    def where(self, *where: Combinable):
        for arg in where:
            if isinstance(arg, bool):
                raise ValueError(
                    "A boolean value has been passed in to a where clause. "
                    "This is probably a mistake. For example "
                    "`.where(MyTable.some_column is None)` instead of "
                    "`.where(MyTable.some_column.is_null())`."
                )

            self._where = And(self._where, arg) if self._where else arg


@dataclass
class OrderByDelegate:

    _order_by: OrderBy = field(default_factory=OrderBy)

    def get_order_by_columns(self) -> t.List[Column]:
        """
        Used to work out which columns are needed for joins.
        """
        return [
            i
            for i in itertools.chain(
                *[i.columns for i in self._order_by.order_by_items]
            )
            if isinstance(i, Column)
        ]

    def order_by(self, *columns: t.Union[Column, OrderByRaw], ascending=True):
        if len(columns) < 1:
            raise ValueError("At least one column must be passed to order_by.")

        self._order_by.order_by_items.append(
            OrderByItem(columns=columns, ascending=ascending)
        )


@dataclass
class LimitDelegate:

    _limit: t.Optional[Limit] = None
    _first: bool = False

    def limit(self, number: int):
        self._limit = Limit(number)

    def copy(self) -> LimitDelegate:
        _limit = self._limit.copy() if self._limit is not None else None
        return self.__class__(_limit=_limit, _first=self._first)


@dataclass
class AsOfDelegate:
    """
    Time travel queries using "As Of" syntax.
    Currently supports Cockroach using AS OF SYSTEM TIME.
    """

    _as_of: t.Optional[AsOf] = None

    def as_of(self, interval: str = "-1s"):
        self._as_of = AsOf(interval)


@dataclass
class DistinctDelegate:

    _distinct: Distinct = field(
        default_factory=lambda: Distinct(enabled=False, on=None)
    )

    def distinct(
        self, enabled: bool, on: t.Optional[t.Sequence[Column]] = None
    ):
        if on and not isinstance(on, collections.abc.Sequence):
            # Check a sequence is passed in, otherwise the user will get some
            # unuseful errors later on.
            raise ValueError("`on` must be a sequence of `Column` instances")

        self._distinct = Distinct(enabled=enabled, on=on)


@dataclass
class ReturningDelegate:
    _returning: t.Optional[Returning] = None

    def returning(self, columns: t.Sequence[Column]):
        self._returning = Returning(columns=list(columns))


@dataclass
class CountDelegate:

    _count: bool = False

    def count(self):
        self._count = True


@dataclass
class AddDelegate:

    _add: t.List[Table] = field(default_factory=list)

    def add(self, *instances: Table, table_class: t.Type[Table]):
        for instance in instances:
            if not isinstance(instance, table_class):
                raise TypeError("Incompatible type added.")

        self._add += instances


@dataclass
class OutputDelegate:
    """
    Example usage:

    .output(as_list=True)
    .output(as_json=True)
    .output(as_json=True, as_list=True)
    """

    _output: Output = field(default_factory=Output)

    def output(
        self,
        as_list: t.Optional[bool] = None,
        as_json: t.Optional[bool] = None,
        load_json: t.Optional[bool] = None,
        nested: t.Optional[bool] = None,
    ):
        """
        :param as_list:
            If each row only returns a single value, compile all of the results
            into a single list.
        :param as_json:
            The results are serialised into JSON. It's equivalent to running
            `json.dumps` on the result.
        :param load_json:
            If True, any JSON fields will have the JSON values returned from
            the database loaded as Python objects.
        """
        # We do it like this, so output can be called multiple times, without
        # overriding any existing values if they're not specified.
        if as_list is not None:
            self._output.as_list = bool(as_list)

        if as_json is not None:
            self._output.as_json = bool(as_json)

        if load_json is not None:
            self._output.load_json = bool(load_json)

        if nested is not None:
            self._output.nested = bool(nested)

    def copy(self) -> OutputDelegate:
        _output = self._output.copy() if self._output is not None else None
        return self.__class__(_output=_output)


@dataclass
class CallbackDelegate:
    """
    Example usage:

    .callback(my_handler_function)
    .callback(print, on=CallbackType.success)
    .callback(my_handler_coroutine)
    .callback([handler1, handler2])
    """

    _callbacks: t.Dict[CallbackType, t.List[Callback]] = field(
        default_factory=lambda: {kind: [] for kind in CallbackType}
    )

    def callback(
        self,
        callbacks: t.Union[t.Callable, t.List[t.Callable]],
        *,
        on: CallbackType,
    ):
        if isinstance(callbacks, list):
            self._callbacks[on].extend(
                Callback(kind=on, target=callback) for callback in callbacks
            )
        else:
            self._callbacks[on].append(Callback(kind=on, target=callbacks))

    async def invoke(self, results: t.Any, *, kind: CallbackType) -> t.Any:
        """
        Utility function that invokes the registered callbacks in the correct
        way, handling both sync and async callbacks. Only callbacks of the
        given kind are invoked.
        Results are passed through the callbacks in the order they were added,
        with each callback able to transform them. This function returns the
        transformed results.
        """
        for callback in self._callbacks[kind]:
            if asyncio.iscoroutinefunction(callback.target):
                results = await callback.target(results)
            else:
                results = callback.target(results)

        return results


@dataclass
class PrefetchDelegate:
    """
    Example usage:

    .prefetch(MyTable.column_a, MyTable.column_b)
    """

    fk_columns: t.List[ForeignKey] = field(default_factory=list)

    def prefetch(self, *fk_columns: t.Union[ForeignKey, t.List[ForeignKey]]):
        """
        :param columns:
            We accept ``ForeignKey`` and ``List[ForeignKey]`` here, in case
            someone passes in a list by accident when using ``all_related()``,
            in which case we flatten the list.

        """
        _fk_columns: t.List[ForeignKey] = []
        for column in fk_columns:
            if isinstance(column, list):
                _fk_columns.extend(column)
            else:
                _fk_columns.append(column)

        combined = self.fk_columns + _fk_columns
        self.fk_columns = combined


@dataclass
class ColumnsDelegate:
    """
    Example usage:

    .columns(MyTable.column_a, MyTable.column_b)
    """

    selected_columns: t.Sequence[Selectable] = field(default_factory=list)

    def columns(self, *columns: t.Union[Selectable, t.List[Selectable]]):
        """
        :param columns:
            We accept ``Selectable`` and ``List[Selectable]`` here, in case
            someone passes in a list by accident when using ``all_columns()``,
            in which case we flatten the list.

        """
        _columns = flatten(columns)
        combined = list(self.selected_columns) + _columns
        self.selected_columns = combined

    def remove_secret_columns(self):
        non_secret = [
            i
            for i in self.selected_columns
            if not isinstance(i, Column) or not i._meta.secret
        ]

        self.selected_columns = non_secret


@dataclass
class ValuesDelegate:
    """
    Used to specify new column values - primarily used in update queries.
    """

    table: t.Type[Table]
    _values: t.Dict[Column, t.Any] = field(default_factory=dict)

    def values(self, values: t.Dict[t.Union[Column, str], t.Any]):
        """
        Example usage:

        .. code-block:: python

            .values({MyTable.column_a: 1})

            # Or:
            .values({'column_a': 1})

            # Or:
            .values(column_a=1})

        """
        cleaned_values: t.Dict[Column, t.Any] = {}
        for key, value in values.items():
            if isinstance(key, Column):
                column = key
            elif isinstance(key, str):
                column = self.table._meta.get_column_by_name(key)
            else:
                raise ValueError(
                    f"Unrecognised key - {key} is neither a Column or the "
                    "name of a Column."
                )
            cleaned_values[column] = value

        self._values.update(cleaned_values)

    def get_sql_values(self) -> t.List[t.Any]:
        """
        Convert any Enums into values, and serialise any JSON.
        """
        return [
            convert_to_sql_value(value=value, column=column)
            for column, value in self._values.items()
        ]


@dataclass
class OffsetDelegate:
    """
    Used to offset the results - for example, to return row 100 and onward.

    Typically used in conjunction with order_by and limit.

    Example usage::

        .offset(100)

    """

    _offset: t.Optional[Offset] = None

    def offset(self, number: int = 0):
        self._offset = Offset(number)


@dataclass
class GroupBy:
    __slots__ = ("columns",)

    columns: t.Sequence[Column]

    @property
    def querystring(self) -> QueryString:
        columns_names = ", ".join(
            i._meta.get_full_name(with_alias=False) for i in self.columns
        )

        return QueryString(f" GROUP BY {columns_names}")

    def __str__(self):
        return self.querystring.__str__()


@dataclass
class GroupByDelegate:
    """
    Used to group results - needed when doing aggregation::

        .group_by(Band.name)

    """

    _group_by: t.Optional[GroupBy] = None

    def group_by(self, *columns: Column):
        self._group_by = GroupBy(columns=columns)


class OnConflictAction(str, Enum):
    """
    Specify which action to take on conflict.
    """

    do_nothing = "DO NOTHING"
    do_update = "DO UPDATE"


@dataclass
class OnConflictItem:
    target: t.Optional[t.Union[str, Column, t.Tuple[Column, ...]]] = None
    action: t.Optional[OnConflictAction] = None
    values: t.Optional[
        t.Sequence[t.Union[Column, t.Tuple[Column, t.Any]]]
    ] = None
    where: t.Optional[Combinable] = None

    @property
    def target_string(self) -> str:
        target = self.target
        assert target

        def to_string(value) -> str:
            if isinstance(value, Column):
                return f'"{value._meta.db_column_name}"'
            else:
                raise ValueError("OnConflict.target isn't a valid type")

        if isinstance(target, str):
            return f'ON CONSTRAINT "{target}"'
        elif isinstance(target, Column):
            return f"({to_string(target)})"
        elif isinstance(target, tuple):
            columns_str = ", ".join([to_string(i) for i in target])
            return f"({columns_str})"
        else:
            raise ValueError("OnConflict.target isn't a valid type")

    @property
    def action_string(self) -> QueryString:
        action = self.action
        if isinstance(action, OnConflictAction):
            if action == OnConflictAction.do_nothing:
                return QueryString(OnConflictAction.do_nothing.value)
            elif action == OnConflictAction.do_update:
                values = []
                query = f"{OnConflictAction.do_update.value} SET"

                if not self.values:
                    raise ValueError("No values specified for `on conflict`")

                for value in self.values:
                    if isinstance(value, Column):
                        column_name = value._meta.db_column_name
                        query += f' "{column_name}"=EXCLUDED."{column_name}",'
                    elif isinstance(value, tuple):
                        column = value[0]
                        value_ = value[1]
                        if isinstance(column, Column):
                            column_name = column._meta.db_column_name
                        else:
                            raise ValueError("Unsupported column type")

                        query += f' "{column_name}"={{}},'
                        values.append(value_)

                return QueryString(query.rstrip(","), *values)

        raise ValueError("OnConflict.action isn't a valid type")

    @property
    def querystring(self) -> QueryString:
        query = " ON CONFLICT"
        values = []

        if self.target:
            query += f" {self.target_string}"

        if self.action:
            query += " {}"
            values.append(self.action_string)

        if self.where:
            query += " WHERE {}"
            values.append(self.where.querystring)

        return QueryString(query, *values)

    def __str__(self) -> str:
        return self.querystring.__str__()


@dataclass
class OnConflict:
    """
    Multiple `ON CONFLICT` statements are allowed - which is why we have this
    parent class.
    """

    on_conflict_items: t.List[OnConflictItem] = field(default_factory=list)

    @property
    def querystring(self) -> QueryString:
        query = "".join("{}" for i in self.on_conflict_items)
        return QueryString(
            query, *[i.querystring for i in self.on_conflict_items]
        )

    def __str__(self) -> str:
        return self.querystring.__str__()


@dataclass
class OnConflictDelegate:
    """
    Used with insert queries to specify what to do when a query fails due to
    a constraint::

        .on_conflict(action='DO NOTHING')

        .on_conflict(action='DO UPDATE', values=[Band.popularity])

        .on_conflict(action='DO UPDATE', values=[(Band.popularity, 1)])

    """

    _on_conflict: OnConflict = field(default_factory=OnConflict)

    def on_conflict(
        self,
        target: t.Optional[t.Union[str, Column, t.Tuple[Column, ...]]] = None,
        action: t.Union[
            OnConflictAction, Literal["DO NOTHING", "DO UPDATE"]
        ] = OnConflictAction.do_nothing,
        values: t.Optional[
            t.Sequence[t.Union[Column, t.Tuple[Column, t.Any]]]
        ] = None,
        where: t.Optional[Combinable] = None,
    ):
        action_: OnConflictAction
        if isinstance(action, OnConflictAction):
            action_ = action
        elif isinstance(action, str):
            action_ = OnConflictAction(action.upper())
        else:
            raise ValueError("Unrecognised `on conflict` action.")

        if where and action_ == OnConflictAction.do_nothing:
            raise ValueError(
                "The `where` option can only be used with DO NOTHING."
            )

        self._on_conflict.on_conflict_items.append(
            OnConflictItem(
                target=target, action=action_, values=values, where=where
            )
        )



from piccolo.conf.apps import AppConfig

from .commands.dump import dump
from .commands.load import load

APP_CONFIG = AppConfig(
    app_name="fixtures",
    migrations_folder_path="",
    table_classes=[],
    migration_dependencies=[],
    commands=[dump, load],
)



from .base import Column, ForeignKeyMeta, OnDelete, OnUpdate, Selectable
from .column_types import (
    JSON,
    JSONB,
    UUID,
    Array,
    BigInt,
    BigSerial,
    Boolean,
    Bytea,
    Date,
    Decimal,
    DoublePrecision,
    Email,
    Float,
    ForeignKey,
    Integer,
    Interval,
    Numeric,
    PrimaryKey,
    Real,
    Secret,
    Serial,
    SmallInt,
    Text,
    Time,
    Timestamp,
    Timestamptz,
    Varchar,
)
from .combination import And, Or, Where
from .m2m import M2M
from .reference import LazyTableReference

__all__ = [
    "Column",
    "ForeignKeyMeta",
    "OnDelete",
    "OnUpdate",
    "Selectable",
    "JSON",
    "JSONB",
    "UUID",
    "Array",
    "BigInt",
    "BigSerial",
    "Boolean",
    "Bytea",
    "Date",
    "Decimal",
    "DoublePrecision",
    "Email",
    "Float",
    "ForeignKey",
    "Integer",
    "Interval",
    "Numeric",
    "PrimaryKey",
    "Real",
    "Secret",
    "Serial",
    "SmallInt",
    "Text",
    "Time",
    "Timestamp",
    "Timestamptz",
    "Varchar",
    "And",
    "Or",
    "Where",
    "M2M",
    "LazyTableReference",
]



from unittest import TestCase

from piccolo.columns import Varchar
from piccolo.schema import SchemaManager
from piccolo.table import Table
from tests.base import engines_skip
from tests.example_apps.music.tables import Manager


class TestTableExists(TestCase):
    def setUp(self):
        Manager.create_table().run_sync()

    def tearDown(self):
        Manager.alter().drop_table().run_sync()

    def test_table_exists(self):
        response = Manager.table_exists().run_sync()
        self.assertTrue(response)


class Band(Table, schema="schema_1"):
    name = Varchar()


@engines_skip("sqlite")
class TestTableExistsSchema(TestCase):
    def setUp(self):
        Band.create_table(auto_create_schema=True).run_sync()

    def tearDown(self):
        SchemaManager().drop_schema(
            "schema_1", if_exists=True, cascade=True
        ).run_sync()

    def test_table_exists(self):
        """
        Make sure it works correctly if the table is in a Postgres schema.
        """
        response = Band.table_exists().run_sync()
        self.assertTrue(response)



import datetime
from unittest import TestCase

import pytest

from piccolo.apps.user.tables import BaseUser
from piccolo.columns import Date, Varchar
from piccolo.columns.combination import WhereRaw
from piccolo.query import OrderByRaw
from piccolo.query.methods.select import Avg, Count, Max, Min, SelectRaw, Sum
from piccolo.query.mixins import DistinctOnError
from piccolo.table import Table, create_db_tables_sync, drop_db_tables_sync
from tests.base import (
    DBTestCase,
    engine_is,
    engine_version_lt,
    engines_only,
    engines_skip,
    is_running_cockroach,
    is_running_sqlite,
    sqlite_only,
)
from tests.example_apps.music.tables import Band, Concert, Manager, Venue


class TestSelect(DBTestCase):
    def test_query_all_columns(self):
        self.insert_row()

        response = Band.select().run_sync()

        if engine_is("cockroach"):
            self.assertDictEqual(
                response[0],
                {
                    "id": response[0]["id"],
                    "name": "Pythonistas",
                    "manager": response[0]["manager"],
                    "popularity": 1000,
                },
            )
        else:
            self.assertDictEqual(
                response[0],
                {
                    "id": 1,
                    "name": "Pythonistas",
                    "manager": 1,
                    "popularity": 1000,
                },
            )

    def test_query_some_columns(self):
        self.insert_row()

        response = Band.select(Band.name).run_sync()

        self.assertDictEqual(response[0], {"name": "Pythonistas"})

    def test_where_equals(self):
        self.insert_row()

        manager = Manager.objects().first().run_sync()

        # This is the recommended way of running these types of queries:
        response = (
            Band.select(Band.name)
            .where(
                Band.manager.id
                == getattr(manager, Band._meta.primary_key._meta.name)
            )
            .run_sync()
        )
        self.assertEqual(response, [{"name": "Pythonistas"}])

        # Other cases which should work:
        response = (
            Band.select(Band.name)
            .where(
                Band.manager
                == getattr(manager, Manager._meta.primary_key._meta.name)
            )
            .run_sync()
        )
        self.assertEqual(response, [{"name": "Pythonistas"}])

        response = (
            Band.select(Band.name)
            .where(
                Band.manager.id
                == getattr(manager, Manager._meta.primary_key._meta.name)
            )
            .run_sync()
        )
        self.assertEqual(response, [{"name": "Pythonistas"}])

        # check multiple arguments inside WHERE clause
        response = (
            Band.select(Band.name)
            .where(Band.manager.id == 1, Band.popularity == 500)
            .run_sync()
        )
        self.assertEqual(response, [])

        # check empty WHERE clause
        response = Band.select(Band.name).where().run_sync()
        self.assertEqual(response, [{"name": "Pythonistas"}])

    @engines_only("postgres", "cockroach")
    def test_where_like_postgres(self):
        """
        Postgres' LIKE is case sensitive.
        """
        self.insert_rows()

        for like_query in ("Python%", "Pythonistas", "%istas", "%ist%"):
            response = (
                Band.select(Band.name)
                .where(Band.name.like(like_query))
                .run_sync()
            )

            self.assertEqual(response, [{"name": "Pythonistas"}])

        for like_query in (
            "PyThonISTAs",
            "PYth%",
            "%ISTAS",
            "%Ist%",
            "PYTHONISTAS",
        ):
            response = (
                Band.select(Band.name)
                .where(Band.name.like(like_query))
                .run_sync()
            )

            self.assertEqual(response, [])

    @sqlite_only
    def test_where_like_sqlite(self):
        """
        SQLite's LIKE is case insensitive for ASCII characters,
        i.e. a == 'A' -> True.
        """
        self.insert_rows()

        for like_query in (
            "Python%",
            "Pythonistas",
            "%istas",
            "%ist%",
            "PYTHONISTAS",
        ):
            response = (
                Band.select(Band.name)
                .where(Band.name.like(like_query))
                .run_sync()
            )

            self.assertEqual(response, [{"name": "Pythonistas"}])

        for like_query in (
            "xyz",
            "XYZ%",
            "%xyz",
            "%xyz%",
        ):
            response = (
                Band.select(Band.name)
                .where(Band.name.like(like_query))
                .run_sync()
            )

            self.assertEqual(response, [])

    @sqlite_only
    def test_where_ilike_sqlite(self):
        """
        SQLite doesn't support ILIKE, so it's just a proxy to LIKE. We still
        have a test to make sure it proxies correctly.
        """
        self.insert_rows()

        for ilike_query in (
            "Python%",
            "Pythonistas",
            "pythonistas",
            "PytHonIStas",
            "%istas",
            "%ist%",
            "%IST%",
        ):
            self.assertEqual(
                Band.select(Band.name)
                .where(Band.name.ilike(ilike_query))
                .run_sync(),
                Band.select(Band.name)
                .where(Band.name.like(ilike_query))
                .run_sync(),
            )

    @engines_only("postgres", "cockroach")
    def test_where_ilike_postgres(self):
        """
        Only Postgres has ILIKE - it's not in the SQL standard. It's for
        case insensitive matching, i.e. 'Foo' == 'foo' -> True.
        """
        self.insert_rows()

        for ilike_query in (
            "Python%",
            "Pythonistas",
            "pythonistas",
            "PytHonIStas",
            "%istas",
            "%ist%",
            "%IST%",
        ):
            response = (
                Band.select(Band.name)
                .where(Band.name.ilike(ilike_query))
                .run_sync()
            )

            self.assertEqual(response, [{"name": "Pythonistas"}])

        for ilike_query in ("Pythonistas1", "%123", "%xyz%"):
            response = (
                Band.select(Band.name)
                .where(Band.name.ilike(ilike_query))
                .run_sync()
            )

            self.assertEqual(response, [])

    def test_where_not_like(self):
        self.insert_rows()

        response = (
            Band.select(Band.name)
            .where(Band.name.not_like("Python%"))
            .order_by(Band.name)
            .run_sync()
        )

        self.assertEqual(
            response, [{"name": "CSharps"}, {"name": "Rustaceans"}]
        )

    def test_where_greater_than(self):
        self.insert_rows()

        response = (
            Band.select(Band.name).where(Band.popularity > 1000).run_sync()
        )

        self.assertEqual(response, [{"name": "Rustaceans"}])

    def test_where_is_null(self):
        self.insert_rows()

        Band(name="Managerless", popularity=0, manager=None).save().run_sync()

        queries = (
            Band.select(Band.name).where(Band.manager == None),  # noqa
            Band.select(Band.name).where(Band.manager.is_null()),
        )

        for query in queries:
            response = query.run_sync()
            self.assertEqual(response, [{"name": "Managerless"}])

    def test_where_bool(self):
        """
        If passing a boolean into a where clause, an exception should be
        raised. It's possible for a user to do this by accident, for example
        ``where(Band.has_drummer is None)``, which evaluates to a boolean.
        """
        with self.assertRaises(ValueError):
            Band.select().where(False)

    def test_where_is_not_null(self):
        self.insert_rows()

        Band(name="Managerless", popularity=0, manager=None).save().run_sync()

        queries = (
            Band.select(Band.name).where(Band.manager != None),  # noqa
            Band.select(Band.name).where(Band.manager.is_not_null()),
        )

        for query in queries:
            response = query.run_sync()
            self.assertEqual(
                response,
                [
                    {"name": "Pythonistas"},
                    {"name": "Rustaceans"},
                    {"name": "CSharps"},
                ],
            )

    def test_where_greater_equal_than(self):
        self.insert_rows()

        response = (
            Band.select(Band.name)
            .where(Band.popularity >= 1000)
            .order_by(Band.name)
            .run_sync()
        )

        self.assertEqual(
            response, [{"name": "Pythonistas"}, {"name": "Rustaceans"}]
        )

    def test_where_less_than(self):
        self.insert_rows()

        response = (
            Band.select(Band.name).where(Band.popularity < 1000).run_sync()
        )

        self.assertEqual(response, [{"name": "CSharps"}])

    def test_where_less_equal_than(self):
        self.insert_rows()

        response = (
            Band.select(Band.name).where(Band.popularity <= 1000).run_sync()
        )

        self.assertEqual(
            response, [{"name": "Pythonistas"}, {"name": "CSharps"}]
        )

    def test_where_raw(self):
        """
        Make sure raw SQL passed in to a where clause works as expected.
        """
        self.insert_rows()

        response = (
            Band.select(Band.name)
            .where(WhereRaw("name = 'Pythonistas'"))
            .run_sync()
        )

        self.assertEqual(response, [{"name": "Pythonistas"}])

    def test_where_raw_with_args(self):
        """
        Make sure raw SQL with args, passed in to a where clause, works
        as expected.
        """
        self.insert_rows()

        response = (
            Band.select(Band.name)
            .where(WhereRaw("name = {}", "Pythonistas"))
            .run_sync()
        )

        self.assertEqual(response, [{"name": "Pythonistas"}])

    def test_where_raw_combined_with_where(self):
        """
        Make sure WhereRaw can be combined with Where.
        """
        self.insert_rows()

        response = (
            Band.select(Band.name)
            .where(
                WhereRaw("name = 'Pythonistas'") | (Band.name == "Rustaceans")
            )
            .run_sync()
        )

        self.assertEqual(
            response, [{"name": "Pythonistas"}, {"name": "Rustaceans"}]
        )

    def test_where_and(self):
        self.insert_rows()

        response = (
            Band.select(Band.name)
            .where((Band.popularity <= 1000) & (Band.name.like("Python%")))
            .run_sync()
        )

        self.assertEqual(response, [{"name": "Pythonistas"}])

    def test_where_or(self):
        self.insert_rows()

        response = (
            Band.select(Band.name)
            .where((Band.name == "Rustaceans") | (Band.name == "CSharps"))
            .order_by(Band.name)
            .run_sync()
        )

        self.assertEqual(
            response, [{"name": "CSharps"}, {"name": "Rustaceans"}]
        )

    @engines_skip("cockroach")
    def test_multiple_where(self):
        """
        Test that chaining multiple where clauses works results in an AND.
        """
        self.insert_rows()

        query = (
            Band.select(Band.name)
            .where(Band.name == "Rustaceans")
            .where(Band.manager == 2)
        )

        response = query.run_sync()

        self.assertEqual(response, [{"name": "Rustaceans"}])
        self.assertIn("AND", query.__str__())

    @engines_skip("cockroach")
    def test_complex_where(self):
        """
        Test a complex where clause - combining AND, and OR.
        """
        self.insert_rows()

        query = (
            Band.select(Band.name)
            .where(
                ((Band.popularity == 2000) & (Band.manager == 2))
                | ((Band.popularity == 10) & (Band.manager == 3))
            )
            .order_by(Band.name)
        )

        response = query.run_sync()

        self.assertEqual(
            response, [{"name": "CSharps"}, {"name": "Rustaceans"}]
        )

    def test_limit(self):
        self.insert_rows()

        response = (
            Band.select(Band.name).order_by(Band.name).limit(1).run_sync()
        )

        self.assertEqual(response, [{"name": "CSharps"}])

    @engines_only("postgres", "cockroach")
    def test_offset_postgres(self):
        self.insert_rows()

        response = (
            Band.select(Band.name).order_by(Band.name).offset(1).run_sync()
        )

        self.assertEqual(
            response, [{"name": "Pythonistas"}, {"name": "Rustaceans"}]
        )

    @sqlite_only
    def test_offset_sqlite(self):
        """
        SQLite requires a limit clause for offset to work.
        """
        self.insert_rows()

        query = Band.select(Band.name).order_by(Band.name).offset(1)

        with self.assertRaises(ValueError):
            query.run_sync()

        query = query.limit(5)
        response = query.run_sync()

        self.assertEqual(
            response, [{"name": "Pythonistas"}, {"name": "Rustaceans"}]
        )

    def test_first(self):
        self.insert_rows()

        response = (
            Band.select(Band.name).order_by(Band.name).first().run_sync()
        )

        self.assertEqual(response, {"name": "CSharps"})

    def test_count(self):
        self.insert_rows()

        response = Band.count().where(Band.name == "Pythonistas").run_sync()

        self.assertEqual(response, 1)

    def test_distinct(self):
        """
        Make sure the distinct clause works.
        """
        self.insert_rows()
        self.insert_rows()

        query = Band.select(Band.name).where(Band.name == "Pythonistas")
        self.assertNotIn("DISTINCT", query.__str__())

        response = query.run_sync()
        self.assertEqual(
            response, [{"name": "Pythonistas"}, {"name": "Pythonistas"}]
        )

        query = query.distinct()
        self.assertIn("DISTINCT", query.__str__())

        response = query.run_sync()
        self.assertEqual(response, [{"name": "Pythonistas"}])

    def test_distinct_on(self):
        """
        Make sure the distinct clause works, with the ``on`` param.
        """
        self.insert_rows()
        self.insert_rows()

        query = Band.select(Band.name).where(Band.name == "Pythonistas")
        self.assertNotIn("DISTINCT", query.__str__())

        response = query.run_sync()
        self.assertEqual(
            response, [{"name": "Pythonistas"}, {"name": "Pythonistas"}]
        )

        query = query.distinct()
        self.assertIn("DISTINCT", query.__str__())

        response = query.run_sync()
        self.assertEqual(response, [{"name": "Pythonistas"}])

    def test_count_group_by(self):
        """
        Test grouping and counting all rows.
        """
        self.insert_rows()
        self.insert_rows()

        response = (
            Band.select(Band.name, Count())
            .group_by(Band.name)
            .order_by(Band.name)
            .run_sync()
        )

        self.assertEqual(
            response,
            [
                {"name": "CSharps", "count": 2},
                {"name": "Pythonistas", "count": 2},
                {"name": "Rustaceans", "count": 2},
            ],
        )

    def test_count_with_alias_group_by(self):
        """
        Test grouping and counting all rows with alias.
        """
        self.insert_rows()
        self.insert_rows()

        response = (
            Band.select(Band.name, Count(alias="total"))
            .group_by(Band.name)
            .order_by(Band.name)
            .run_sync()
        )

        self.assertEqual(
            response,
            [
                {"name": "CSharps", "total": 2},
                {"name": "Pythonistas", "total": 2},
                {"name": "Rustaceans", "total": 2},
            ],
        )

    def test_count_with_as_alias_group_by(self):
        """
        Test grouping and counting all rows with as_alias.
        """
        self.insert_rows()
        self.insert_rows()

        response = (
            Band.select(Band.name, Count().as_alias("total"))
            .group_by(Band.name)
            .order_by(Band.name)
            .run_sync()
        )

        self.assertEqual(
            response,
            [
                {"name": "CSharps", "total": 2},
                {"name": "Pythonistas", "total": 2},
                {"name": "Rustaceans", "total": 2},
            ],
        )

    def test_count_column_group_by(self):
        """
        Test grouping and counting a specific column. Any null values in the
        specified column will be omitted from the count.
        """
        self.insert_rows()
        self.insert_rows()
        self.run_sync(
            """
            INSERT INTO band (
                name,
                manager,
                popularity
            ) VALUES (
                'SomeBand',
                null,
                1000
            );"""
        )

        response = (
            Band.select(Band.manager.name, Count(Band.manager))
            .group_by(Band.manager.name)
            .order_by(Band.manager.name)
            .run_sync()
        )

        # We need to sort them, because SQLite and Postgres treat Null
        # differently when sorting.
        response = sorted(response, key=lambda x: x["manager.name"] or "")

        self.assertEqual(
            response,
            [
                {"manager.name": None, "count": 0},
                {"manager.name": "Graydon", "count": 2},
                {"manager.name": "Guido", "count": 2},
                {"manager.name": "Mads", "count": 2},
            ],
        )

        # This time the nulls should be counted, as we omit the column argument
        # from Count:
        response = (
            Band.select(Band.manager.name, Count())
            .group_by(Band.manager.name)
            .order_by(Band.manager.name)
            .run_sync()
        )

        response = sorted(response, key=lambda x: x["manager.name"] or "")

        self.assertEqual(
            response,
            [
                {"manager.name": None, "count": 1},
                {"manager.name": "Graydon", "count": 2},
                {"manager.name": "Guido", "count": 2},
                {"manager.name": "Mads", "count": 2},
            ],
        )

    def test_avg(self):
        self.insert_rows()

        response = Band.select(Avg(Band.popularity)).first().run_sync()

        self.assertEqual(float(response["avg"]), 1003.3333333333334)

    def test_avg_alias(self):
        self.insert_rows()

        response = (
            Band.select(Avg(Band.popularity, alias="popularity_avg"))
            .first()
            .run_sync()
        )

        self.assertEqual(float(response["popularity_avg"]), 1003.3333333333334)

    def test_avg_as_alias_method(self):
        self.insert_rows()

        response = (
            Band.select(Avg(Band.popularity).as_alias("popularity_avg"))
            .first()
            .run_sync()
        )

        self.assertEqual(float(response["popularity_avg"]), 1003.3333333333334)

    def test_avg_with_where_clause(self):
        self.insert_rows()

        response = (
            Band.select(Avg(Band.popularity))
            .where(Band.popularity > 500)
            .first()
            .run_sync()
        )

        self.assertEqual(response["avg"], 1500)

    def test_avg_alias_with_where_clause(self):
        """
        Make sure we can specify aliases for the aggregate function,
        when the column is also being used in a where clause.
        """
        self.insert_rows()

        response = (
            Band.select(Avg(Band.popularity, alias="popularity_avg"))
            .where(Band.popularity > 500)
            .first()
            .run_sync()
        )

        self.assertEqual(response["popularity_avg"], 1500)

    def test_avg_as_alias_method_with_where_clause(self):
        """
        Make sure we can specify as_alias method for the aggregate function,
        when the column is also being used in a where clause.
        """
        self.insert_rows()

        response = (
            Band.select(Avg(Band.popularity).as_alias("popularity_avg"))
            .where(Band.popularity > 500)
            .first()
            .run_sync()
        )

        self.assertEqual(response["popularity_avg"], 1500)

    def test_max(self):
        self.insert_rows()

        response = Band.select(Max(Band.popularity)).first().run_sync()

        self.assertEqual(response["max"], 2000)

    def test_max_alias(self):
        self.insert_rows()

        response = (
            Band.select(Max(Band.popularity, alias="popularity_max"))
            .first()
            .run_sync()
        )

        self.assertEqual(response["popularity_max"], 2000)

    def test_max_as_alias_method(self):
        self.insert_rows()

        response = (
            Band.select(Max(Band.popularity).as_alias("popularity_max"))
            .first()
            .run_sync()
        )

        self.assertEqual(response["popularity_max"], 2000)

    def test_min(self):
        self.insert_rows()

        response = Band.select(Min(Band.popularity)).first().run_sync()

        self.assertEqual(response["min"], 10)

    def test_min_alias(self):
        self.insert_rows()

        response = (
            Band.select(Min(Band.popularity, alias="popularity_min"))
            .first()
            .run_sync()
        )

        self.assertEqual(response["popularity_min"], 10)

    def test_min_as_alias_method(self):
        self.insert_rows()

        response = (
            Band.select(Min(Band.popularity).as_alias("popularity_min"))
            .first()
            .run_sync()
        )

        self.assertEqual(response["popularity_min"], 10)

    def test_sum(self):
        self.insert_rows()

        response = Band.select(Sum(Band.popularity)).first().run_sync()

        self.assertEqual(response["sum"], 3010)

    def test_sum_alias(self):
        self.insert_rows()

        response = (
            Band.select(Sum(Band.popularity, alias="popularity_sum"))
            .first()
            .run_sync()
        )

        self.assertEqual(response["popularity_sum"], 3010)

    def test_sum_as_alias_method(self):
        self.insert_rows()

        response = (
            Band.select(Sum(Band.popularity).as_alias("popularity_sum"))
            .first()
            .run_sync()
        )

        self.assertEqual(response["popularity_sum"], 3010)

    def test_sum_with_where_clause(self):
        self.insert_rows()

        response = (
            Band.select(Sum(Band.popularity))
            .where(Band.popularity > 500)
            .first()
            .run_sync()
        )

        self.assertEqual(response["sum"], 3000)

    def test_sum_alias_with_where_clause(self):
        """
        Make sure we can specify aliases for the aggregate function,
        when the column is also being used in a where clause.
        """
        self.insert_rows()

        response = (
            Band.select(Sum(Band.popularity, alias="popularity_sum"))
            .where(Band.popularity > 500)
            .first()
            .run_sync()
        )

        self.assertEqual(response["popularity_sum"], 3000)

    def test_sum_as_alias_method_with_where_clause(self):
        """
        Make sure we can specify as_alias method for the aggregate function,
        when the column is also being used in a where clause.
        """
        self.insert_rows()

        response = (
            Band.select(Sum(Band.popularity).as_alias("popularity_sum"))
            .where(Band.popularity > 500)
            .first()
            .run_sync()
        )

        self.assertEqual(response["popularity_sum"], 3000)

    def test_chain_different_functions(self):
        self.insert_rows()

        response = (
            Band.select(Avg(Band.popularity), Sum(Band.popularity))
            .first()
            .run_sync()
        )

        self.assertEqual(float(response["avg"]), 1003.3333333333334)
        self.assertEqual(response["sum"], 3010)

    def test_chain_different_functions_alias(self):
        self.insert_rows()

        response = (
            Band.select(
                Avg(Band.popularity, alias="popularity_avg"),
                Sum(Band.popularity, alias="popularity_sum"),
            )
            .first()
            .run_sync()
        )

        self.assertEqual(float(response["popularity_avg"]), 1003.3333333333334)
        self.assertEqual(response["popularity_sum"], 3010)

    def test_avg_validation(self):
        with self.assertRaises(ValueError):
            Band.select(Avg(Band.name)).run_sync()

    def test_sum_validation(self):
        with self.assertRaises(ValueError):
            Band.select(Sum(Band.name)).run_sync()

    def test_columns(self):
        """
        Make sure the colums method can be used to specify which columns to
        query.
        """
        self.insert_rows()

        response = (
            Band.select()
            .columns(Band.name)
            .where(Band.name == "Pythonistas")
            .first()
            .run_sync()
        )
        self.assertEqual(response, {"name": "Pythonistas"})

        # Multiple calls to 'columns' should be additive.
        response = (
            Band.select()
            .columns(Band._meta.primary_key)
            .columns(Band.name)
            .where(Band.name == "Pythonistas")
            .first()
            .run_sync()
        )

        if engine_is("cockroach"):
            self.assertEqual(
                response, {"id": response["id"], "name": "Pythonistas"}
            )
        else:
            self.assertEqual(response, {"id": 1, "name": "Pythonistas"})

    def test_call_chain(self):
        """
        Make sure the call chain lengths are the correct size.
        """
        self.assertEqual(len(Concert.band_1.name._meta.call_chain), 1)
        self.assertEqual(len(Concert.band_1.manager.name._meta.call_chain), 2)

    def test_as_alias(self):
        """
        Make sure we can specify aliases for the columns.
        """
        self.insert_row()
        response = Band.select(Band.name.as_alias("title")).run_sync()
        self.assertEqual(response, [{"title": "Pythonistas"}])

    def test_as_alias_with_join(self):
        """
        Make sure we can specify aliases for the column, when performing a
        join.
        """
        self.insert_row()
        response = Band.select(
            Band.manager.name.as_alias("manager_name")
        ).run_sync()
        self.assertEqual(response, [{"manager_name": "Guido"}])

    def test_as_alias_with_where_clause(self):
        """
        Make sure we can specify aliases for the column, when the column is
        also being used in a where clause.
        """
        self.insert_row()
        response = (
            Band.select(Band.name, Band.manager.name.as_alias("manager_name"))
            .where(Band.manager.name == "Guido")
            .run_sync()
        )
        self.assertEqual(
            response, [{"name": "Pythonistas", "manager_name": "Guido"}]
        )

    @pytest.mark.skipif(
        is_running_sqlite() and engine_version_lt(3.35),
        reason="SQLite doesn't have math functions in this version.",
    )
    @pytest.mark.skipif(
        is_running_cockroach(),
        reason=(
            "Cockroach raises an error when trying to use the log function."
        ),
    )
    def test_select_raw(self):
        """
        Make sure ``SelectRaw`` can be used in select queries.
        """
        self.insert_row()
        response = Band.select(
            Band.name, SelectRaw("round(log(popularity)) AS popularity_log")
        ).run_sync()
        self.assertListEqual(
            response, [{"name": "Pythonistas", "popularity_log": 3.0}]
        )


class TestSelectSecret(TestCase):
    def setUp(self):
        BaseUser.create_table().run_sync()

    def tearDown(self):
        BaseUser.alter().drop_table().run_sync()

    def test_secret(self):
        """
        Make sure that secret fields are omitted from the response when
        requested.
        """
        user = BaseUser(username="piccolo", password="piccolo123")
        user.save().run_sync()

        user_dict = BaseUser.select(exclude_secrets=True).first().run_sync()
        self.assertNotIn("password", user_dict.keys())


class TestSelectSecretParameter(TestCase):
    def setUp(self):
        Venue.create_table().run_sync()

    def tearDown(self):
        Venue.alter().drop_table().run_sync()

    def test_secret_parameter(self):
        """
        Make sure that fields with parameter ``secret=True`` are omitted
        from the response when requested.
        """
        venue = Venue(name="The Garage", capacity=1000)
        venue.save().run_sync()

        venue_dict = Venue.select(exclude_secrets=True).first().run_sync()
        if engine_is("cockroach"):
            self.assertTrue(
                venue_dict, {"id": venue_dict["id"], "name": "The Garage"}
            )
        else:
            self.assertTrue(venue_dict, {"id": 1, "name": "The Garage"})
        self.assertNotIn("capacity", venue_dict.keys())


class TestSelectOrderBy(TestCase):
    """
    We use TestCase, rather than DBTestCase, as we want a lot of data to test
    with.
    """

    def setUp(self):
        """
        Create tables and lots of test data.
        """
        create_db_tables_sync(Band, Manager)

        data = [
            {
                "band_name": "Pythonistas",
                "manager_name": "Guido",
                "popularity": 1000,
            },
            {
                "band_name": "Rustaceans",
                "manager_name": "Graydon",
                "popularity": 800,
            },
            {
                "band_name": "C-Sharps",
                "manager_name": "Anders",
                "popularity": 800,
            },
            {
                "band_name": "Rubyists",
                "manager_name": "Matz",
                "popularity": 820,
            },
        ]

        for item in data:
            manager = (
                Manager.objects().create(name=item["manager_name"]).run_sync()
            )

            Band.objects().create(
                name=item["band_name"],
                manager=manager,
                popularity=item["popularity"],
            ).run_sync()

    def tearDown(self):
        drop_db_tables_sync(Band, Manager)

    def test_ascending(self):
        response = Band.select(Band.name).order_by(Band.name).run_sync()

        self.assertEqual(
            response,
            [
                {"name": "C-Sharps"},
                {"name": "Pythonistas"},
                {"name": "Rubyists"},
                {"name": "Rustaceans"},
            ],
        )

    def test_descending(self):
        response = (
            Band.select(Band.name)
            .order_by(Band.name, ascending=False)
            .run_sync()
        )

        self.assertEqual(
            response,
            [
                {"name": "Rustaceans"},
                {"name": "Rubyists"},
                {"name": "Pythonistas"},
                {"name": "C-Sharps"},
            ],
        )

    def test_string(self):
        """
        Make sure strings can be used to identify columns if the user prefers.
        """
        response = Band.select(Band.name).order_by("name").run_sync()

        self.assertEqual(
            response,
            [
                {"name": "C-Sharps"},
                {"name": "Pythonistas"},
                {"name": "Rubyists"},
                {"name": "Rustaceans"},
            ],
        )

    def test_string_unrecognised(self):
        """
        Make sure an unrecognised column name raises an Exception.
        """
        with self.assertRaises(ValueError) as manager:
            Band.select(Band.name).order_by("foo")

        self.assertEqual(
            manager.exception.__str__(),
            "No matching column found with name == foo",
        )

    def test_multiple_columns_ascending(self):
        """
        Make sure we can order by multiple columns.
        """
        response = (
            Band.select(Band.popularity, Band.name)
            .order_by(Band.popularity, Band.name)
            .run_sync()
        )

        self.assertEqual(
            response,
            [
                {"popularity": 800, "name": "C-Sharps"},
                {"popularity": 800, "name": "Rustaceans"},
                {"popularity": 820, "name": "Rubyists"},
                {"popularity": 1000, "name": "Pythonistas"},
            ],
        )

    def test_multiple_columns_descending(self):
        """
        Make sure we can order by multiple columns, descending.
        """
        response = (
            Band.select(Band.popularity, Band.name)
            .order_by(Band.popularity, Band.name, ascending=False)
            .run_sync()
        )

        self.assertEqual(
            response,
            [
                {"popularity": 1000, "name": "Pythonistas"},
                {"popularity": 820, "name": "Rubyists"},
                {"popularity": 800, "name": "Rustaceans"},
                {"popularity": 800, "name": "C-Sharps"},
            ],
        )

    def test_join(self):
        """
        Make sure that we can order using columns in related tables.
        """
        response = (
            Band.select(Band.manager.name.as_alias("manager_name"), Band.name)
            .order_by(Band.manager.name)
            .run_sync()
        )
        self.assertEqual(
            response,
            [
                {"manager_name": "Anders", "name": "C-Sharps"},
                {"manager_name": "Graydon", "name": "Rustaceans"},
                {"manager_name": "Guido", "name": "Pythonistas"},
                {"manager_name": "Matz", "name": "Rubyists"},
            ],
        )

    def test_ascending_descending(self):
        """
        Make sure we can combine ascending and descending.
        """
        response = (
            Band.select(Band.popularity, Band.name)
            .order_by(Band.popularity)
            .order_by(Band.name, ascending=False)
            .run_sync()
        )

        self.assertEqual(
            response,
            [
                {"popularity": 800, "name": "Rustaceans"},
                {"popularity": 800, "name": "C-Sharps"},
                {"popularity": 820, "name": "Rubyists"},
                {"popularity": 1000, "name": "Pythonistas"},
            ],
        )

    def test_order_by_raw(self):
        """
        Maker sure ``OrderByRaw`` can be used, to order by anything the user
        wants.
        """
        response = (
            Band.select(Band.name).order_by(OrderByRaw("name")).run_sync()
        )

        self.assertEqual(
            response,
            [
                {"name": "C-Sharps"},
                {"name": "Pythonistas"},
                {"name": "Rubyists"},
                {"name": "Rustaceans"},
            ],
        )


class Album(Table):
    band = Varchar()
    title = Varchar()
    release_date = Date()


class TestDistinctOn(TestCase):
    def setUp(self):
        Album.create_table().run_sync()

    def tearDown(self):
        Album.alter().drop_table().run_sync()

    @engines_only("postgres", "cockroach")
    def test_distinct_on(self):
        """
        Make sure the ``distinct`` method can be used to create a
        ``DISTINCT ON`` clause.
        """
        Album.insert(
            Album(
                {
                    Album.band: "Pythonistas",
                    Album.title: "P1",
                    Album.release_date: datetime.date(
                        year=2022, month=1, day=1
                    ),
                }
            ),
            Album(
                {
                    Album.band: "Pythonistas",
                    Album.title: "P2",
                    Album.release_date: datetime.date(
                        year=2023, month=1, day=1
                    ),
                }
            ),
            Album(
                {
                    Album.band: "Rustaceans",
                    Album.title: "R1",
                    Album.release_date: datetime.date(
                        year=2022, month=1, day=1
                    ),
                }
            ),
            Album(
                {
                    Album.band: "Rustaceans",
                    Album.title: "R2",
                    Album.release_date: datetime.date(
                        year=2023, month=1, day=1
                    ),
                }
            ),
            Album(
                {
                    Album.band: "C-Sharps",
                    Album.title: "C1",
                    Album.release_date: datetime.date(
                        year=2022, month=1, day=1
                    ),
                }
            ),
            Album(
                {
                    Album.band: "C-Sharps",
                    Album.title: "C2",
                    Album.release_date: datetime.date(
                        year=2023, month=1, day=1
                    ),
                }
            ),
        ).run_sync()

        # Get the most recent album for each band.
        query = (
            Album.select(Album.band, Album.title)
            .distinct(on=[Album.band])
            .order_by(Album.band)
            .order_by(Album.release_date, ascending=False)
        )
        self.assertIn("DISTINCT ON", query.__str__())
        response = query.run_sync()

        self.assertEqual(
            response,
            [
                {"band": "C-Sharps", "title": "C2"},
                {"band": "Pythonistas", "title": "P2"},
                {"band": "Rustaceans", "title": "R2"},
            ],
        )

    @engines_only("sqlite")
    def test_distinct_on_sqlite(self):
        """
        SQLite doesn't support ``DISTINCT ON``, so a ``ValueError`` should be
        raised.
        """
        with self.assertRaises(NotImplementedError) as manager:
            Album.select().distinct(on=[Album.band])

        self.assertEqual(
            manager.exception.__str__(),
            "SQLite doesn't support DISTINCT ON",
        )

    @engines_only("postgres", "cockroach")
    def test_distinct_on_error(self):
        """
        If we pass in something other than a sequence of columns, it should
        raise a ValueError.
        """
        with self.assertRaises(ValueError) as manager:
            Album.select().distinct(on=Album.band)

        self.assertEqual(
            manager.exception.__str__(),
            "`on` must be a sequence of `Column` instances",
        )

    @engines_only("postgres", "cockroach")
    def test_distinct_on_order_by_error(self):
        """
        The first column passed to `order_by` must match the first column
        passed to `on`, otherwise an exception is raised.
        """
        with self.assertRaises(DistinctOnError):
            Album.select().distinct(on=[Album.band]).order_by(
                Album.release_date
            ).run_sync()



from __future__ import annotations

import abc
import typing as t

from piccolo.engine.base import Engine
from piccolo.engine.finder import engine_finder
from piccolo.querystring import QueryString
from piccolo.utils.sync import run_sync


class SchemaDDLBase(abc.ABC):

    db: Engine

    @abc.abstractproperty
    def ddl(self) -> str:
        pass

    def __await__(self):
        return self.run().__await__()

    async def run(self, in_pool=True):
        return await self.db.run_ddl(self.ddl, in_pool=in_pool)

    def run_sync(self, *args, **kwargs):
        return run_sync(self.run(*args, **kwargs))

    def __str__(self) -> str:
        return self.ddl.__str__()


class CreateSchema(SchemaDDLBase):
    def __init__(
        self,
        schema_name: str,
        *,
        if_not_exists: bool,
        db: Engine,
    ):
        self.schema_name = schema_name
        self.if_not_exists = if_not_exists
        self.db = db

    async def run(self, *args, **kwargs):
        if self.schema_name == "public" or self.schema_name is None:
            return

        return await super().run(self, *args, **kwargs)

    @property
    def ddl(self) -> str:
        query = "CREATE SCHEMA"
        if self.if_not_exists:
            query += " IF NOT EXISTS"
        query += f' "{self.schema_name}"'

        return query


class DropSchema(SchemaDDLBase):
    def __init__(
        self,
        schema_name: str,
        *,
        if_exists: bool,
        cascade: bool,
        db: Engine,
    ):
        self.schema_name = schema_name
        self.if_exists = if_exists
        self.cascade = cascade
        self.db = db

    @property
    def ddl(self) -> str:
        query = "DROP SCHEMA"
        if self.if_exists:
            query += " IF EXISTS"
        query += f' "{self.schema_name}"'

        if self.cascade:
            query += " CASCADE"

        return query


class RenameSchema(SchemaDDLBase):
    def __init__(
        self,
        schema_name: str,
        new_schema_name: str,
        db: Engine,
    ):
        self.schema_name = schema_name
        self.new_schema_name = new_schema_name
        self.db = db

    @property
    def ddl(self):
        return (
            f'ALTER SCHEMA "{self.schema_name}" '
            f'RENAME TO "{self.new_schema_name}"'
        )


class MoveTable(SchemaDDLBase):
    def __init__(
        self,
        table_name: str,
        new_schema: str,
        db: Engine,
        current_schema: t.Optional[str] = None,
    ):
        self.table_name = table_name
        self.current_schema = current_schema
        self.new_schema = new_schema
        self.db = db

    @property
    def ddl(self):
        table_name = f'"{self.table_name}"'
        if self.current_schema:
            table_name = f'"{self.current_schema}".{table_name}'

        return f'ALTER TABLE {table_name} SET SCHEMA "{self.new_schema}"'


class ListTables:
    def __init__(self, db: Engine, schema_name: str):
        self.db = db
        self.schema_name = schema_name

    async def run(self):
        response = await self.db.run_querystring(
            QueryString(
                """
                SELECT table_name
                FROM information_schema.tables
                WHERE table_schema = {}
                """,
                self.schema_name,
            )
        )
        return [i["table_name"] for i in response]

    def run_sync(self):
        return run_sync(self.run())

    def __await__(self):
        return self.run().__await__()


class ListSchemas:
    def __init__(self, db: Engine):
        self.db = db

    async def run(self):
        response = await self.db.run_querystring(
            QueryString("SELECT schema_name FROM information_schema.schemata")
        )
        return [i["schema_name"] for i in response]

    def run_sync(self):
        return run_sync(self.run())

    def __await__(self):
        return self.run().__await__()


class SchemaManager:
    def __init__(self, db: t.Optional[Engine] = None):
        """
        A useful utility class for interacting with schemas.

        :param db:
            Used to execute the database queries. If not specified, we try and
            import it from ``piccolo_conf.py``.
        """
        db = db or engine_finder()

        if not db:
            raise ValueError("The DB can't be found.")

        self.db = db

    def create_schema(
        self, schema_name: str, *, if_not_exists: bool = True
    ) -> CreateSchema:
        """
        Creates the specified schema::

            >>> await SchemaManager().create_schema(schema_name="music")

        :param schema_name:
            The name of the schema to create.
        :param if_not_exists:
            No error will be raised if the schema already exists.

        """
        return CreateSchema(
            schema_name=schema_name,
            if_not_exists=if_not_exists,
            db=self.db,
        )

    def drop_schema(
        self,
        schema_name: str,
        *,
        if_exists: bool = True,
        cascade: bool = False,
    ) -> DropSchema:
        """
        Drops the specified schema::

            >>> await SchemaManager().drop_schema(schema_name="music")

        :param schema_name:
            The name of the schema to drop.
        :param if_exists:
            No error will be raised if the schema doesn't exist.
        :param cascade:
            If ``True`` then it will automatically drop the tables within the
            schema.

        """
        return DropSchema(
            schema_name=schema_name,
            if_exists=if_exists,
            cascade=cascade,
            db=self.db,
        )

    def rename_schema(
        self, schema_name: str, new_schema_name: str
    ) -> RenameSchema:
        """
        Rename the schema::

            >>> await SchemaManager().rename_schema(
            ...     schema_name="music",
            ...     new_schema_name="music_info"
            ... )

        :param schema_name:
            The current name of the schema.
        :param new_schema_name:
            What to rename the schema to.

        """
        return RenameSchema(
            schema_name=schema_name,
            new_schema_name=new_schema_name,
            db=self.db,
        )

    def move_table(
        self,
        table_name: str,
        new_schema: str,
        current_schema: t.Optional[str] = None,
    ) -> MoveTable:
        """
        Moves a table to a different schema::

            >>> await SchemaManager().move_schema(
            ...     table_name='my_table',
            ...     new_schema='schema_1'
            ... )

        :param table_name:
            The name of the table to move.
        :param new_schema:
            The name of the scheam you want to move the table too.
        :current_schema:
            If not specified, ``'public'`` is assumed.

        """
        return MoveTable(
            table_name=table_name,
            new_schema=new_schema,
            current_schema=current_schema,
            db=self.db,
        )

    def list_tables(self, schema_name: str) -> ListTables:
        """
        Returns the name of each table in the given schema::

            >>> await SchemaManager().list_tables(schema_name="music")
            ['band', 'manager']

        :param schema_name:
            List the tables in this schema.

        """
        return ListTables(db=self.db, schema_name=schema_name)

    def list_schemas(self) -> ListSchemas:
        """
        Returns the name of each schema in the database::

            >>> await SchemaManager().list_schemas()
            ['public', 'schema_1']

        """
        return ListSchemas(db=self.db)
